/// <reference types="node" />
import { BinaryStream } from "node-opcua-binary-stream";
import { DerivedKeys, PrivateKey } from "node-opcua-crypto";
import { BaseUAObject } from "node-opcua-factory";
import { ExpandedNodeId } from "node-opcua-nodeid";
import { MessageSecurityMode } from "node-opcua-service-secure-channel";
import { StatusCode } from "node-opcua-status-code";
import { MessageBuilderBase, MessageBuilderBaseOptions } from "node-opcua-transport";
import { SecurityHeader } from "./secure_message_chunk_manager";
import { CryptoFactory, SecurityPolicy } from "./security_policy";
export interface SecurityToken {
    tokenId: number;
    expired?: boolean;
    revisedLifetime: number;
}
export interface ObjectFactory {
    constructObject: (binaryEncodingNodeId: ExpandedNodeId) => BaseUAObject;
    hasConstructor: (binaryEncodingNodeId: ExpandedNodeId) => boolean;
}
export interface MessageBuilderOptions extends MessageBuilderBaseOptions {
    securityMode?: MessageSecurityMode;
    privateKey?: PrivateKey;
    objectFactory?: ObjectFactory;
    signatureLength?: number;
    name?: string;
}
export interface SecurityTokenAndDerivedKeys {
    securityToken: SecurityToken;
    derivedKeys: DerivedKeys | null;
}
export declare const invalidPrivateKey: PrivateKey;
type PacketInfo = any;
export interface MessageBuilder extends MessageBuilderBase {
    on(eventName: "startChunk", eventHandler: (info: PacketInfo, data: Buffer) => void): this;
    on(eventName: "chunk", eventHandler: (chunk: Buffer) => void): this;
    on(eventName: "error", eventHandler: (err: Error, statusCode: StatusCode, requestId: number | null) => void): this;
    on(eventName: "full_message_body", eventHandler: (fullMessageBody: Buffer) => void): this;
    on(eventName: "message", eventHandler: (obj: BaseUAObject, msgType: string, requestId: number, channelId: number) => void): this;
    on(eventName: "abandon", eventHandler: (requestId: number) => void): this;
    on(eventName: "invalid_message", eventHandler: (obj: BaseUAObject) => void): this;
    on(eventName: "invalid_sequence_number", eventHandler: (expectedSequenceNumber: number, sequenceNumber: number) => void): this;
    on(eventName: "new_token", eventHandler: (tokenId: number) => void): this;
    emit(eventName: "startChunk", info: PacketInfo, data: Buffer): boolean;
    emit(eventName: "chunk", chunk: Buffer): boolean;
    emit(eventName: "error", err: Error, statusCode: StatusCode, requestId: number | null): boolean;
    emit(eventName: "full_message_body", fullMessageBody: Buffer): boolean;
    emit(eventName: "message", obj: BaseUAObject, msgType: string, requestId: number, channelId: number): boolean;
    emit(eventName: "invalid_message", evobj: BaseUAObject): boolean;
    emit(eventName: "invalid_sequence_number", expectedSequenceNumber: number, sequenceNumber: number): boolean;
    emit(eventName: "new_token", tokenId: number): boolean;
    emit(eventName: "abandon"): boolean;
}
/**
 * @class MessageBuilder
 * @extends MessageBuilderBase
 * @constructor
 *
 * @param options
 * @param options.securityMode {MessageSecurityMode} the security Mode
 * @param [options.objectFactory=factories] a object that provides a constructObject(id) method
 */
export declare class MessageBuilder extends MessageBuilderBase {
    securityPolicy: SecurityPolicy;
    securityMode: MessageSecurityMode;
    cryptoFactory: CryptoFactory | null;
    securityHeader?: SecurityHeader;
    protected id: string;
    private readonly objectFactory;
    private _previousSequenceNumber;
    private _tokenStack;
    private privateKey;
    constructor(options: MessageBuilderOptions);
    setSecurity(securityMode: MessageSecurityMode, securityPolicy: SecurityPolicy): void;
    dispose(): void;
    pushNewToken(securityToken: SecurityToken, derivedKeys: DerivedKeys | null): void;
    protected _read_headers(binaryStream: BinaryStream): boolean;
    protected _decodeMessageBody(fullMessageBody: Buffer): boolean;
    private _validateSequenceNumber;
    private _decrypt_OPN;
    private tokenIds;
    private _select_matching_token;
    private _decrypt_MSG;
    private _decrypt;
    private _safe_decode_message_body;
}
export {};
